/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.graveng.rci.client;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Base64;
import java.util.EventListener;
import java.util.HashMap;
import java.util.LinkedList;

import com.graveng.rci.NullObject;
import com.graveng.rci.Packet;
import com.graveng.rci.TimeoutException;
import com.graveng.rci.crypt.AesCrypt;
import com.graveng.rci.crypt.CryptFilter;
import com.graveng.rci.crypt.HashCrypt;
import com.graveng.rci.util.Logger;

/**
 * RCIクライアントクラス
 * <p>
 * RCIサーバのIPアドレスとポート番号を指定しconnect後、遠隔クラス生成、遠隔メソッド実行を行う。
 * <p>
 * リスナー登録することでサーバからの非同期通知を受けることもできる。
 * <p>
 * メソッドコールは非同期ではないからRCIクライアントは基本的に同期通信である。複数のメソッドコールを
 * スレッドでやりたい場合は、Rciのインスタンスをそれぞれのスレッドで動かすこと。
 * （RciServerはsshと同じように同じユーザでの複数同時接続を禁止していない。）
 *
 */
public class Rci {
	/**
	 * ログファイル名を変更する場合の環境変数名
	 */
	public static final String logFileNameEnv = "com.graveng.rci.util.Logger.logFileName";

	/**
	 * ログレベルを変更する場合の環境変数名
	 */
	public static final String logLevelEnv = "com.graveng.rci.util.Logger.logLevel";

	/**
	 * クライアントソケット
	 */
	protected Socket socket = null;

	/**
	 * １つのコネクションの送受信を行う。
	 */
	protected Invoker invoker = null;

	/**
	 * RCIサーバで発生した例外をシリアライズして受け、その例外を送出する場合はtrueにする。
	 */
	protected boolean isThrowException = true;

	/**
	 * サーバからの非同期イベントの受信処理を有効にする場合はtrueにする。<br>
	 * リスナー登録により非同期イベントを受信しないならfalseにした方がやや効率がよい。<br>
	 */
	protected boolean isUseEventListener = true;

	/**
	 * RCIサーバからの非同期通知を受け取るインスタンスの番号（登録されるごとにインクリメントする）
	 */
	protected int listenerNumber = 0;

	/**
	 * RCIサーバからの非同期通知を受け取るクラスを保持する。key:listenerNumber, value:リスナークラス
	 */
	protected HashMap<Integer, EventListener> listenerMap = new HashMap<Integer, EventListener>();

	/**
	 * デフォルトコンストラクタ、Rciクライアントを生成する
	 */
	public Rci() {
		updateLogSetup();
	}

	/**
	 * サーバからの非同期通知を受けられる状態かを返す。
	 * <p>
	 * @return isUseEventListener 非同期通知を受けられる状態ならtrue
	 */
	public boolean isUseEventListener() {
		return isUseEventListener;
	}

	/**
	 * サーバからの非同期通知を受けられる状態かを設定する。
	 * <p>
	 * @param isUseEventListener 非同期通知を受けられる状態ならtrue。同期通信のみで無限待ちにする場合はfalse
	 */
	public void setUseEventListener(boolean isUseEventListener) {
		this.isUseEventListener = isUseEventListener;
	}

	/**
	 * ログファイル名、ログレベルを環境変数から設定する場合はこのメソッドを呼び出す。
	 */
	protected void updateLogSetup() {
		String logFileName = System.getenv(logFileNameEnv);
		String logLevel = System.getenv(logLevelEnv);
		Logger.setLogFileName(logFileName);
		Logger.setLogLevel(logLevel);
	}

	/**
	 * RCIサーバのソケットに接続する。
	 * <p>
	 * @param server RCIサーバのホスト名かIPアドレス
	 * @param port RCIサーバのポート番号
	 * @throws UnknownHostException ホストが見つからない場合
	 * @throws IOException 何らかの通信に失敗した場合
	 */
	public void connect(String server, int port) throws UnknownHostException, IOException {
		// サーバに接続する
		socket = new Socket(server, port);
		invoker = new Invoker(socket);
		if(isUseEventListener) {
			startReadThread();
		}
		Logger.info("connected server["+server+"] port["+port+"]");
	}

	/**
	 * １つのコネクションの送受信を行うスレッドを起動する。
	 */
	protected void startReadThread() {
      Thread th = new Thread(invoker);
      th.start();
 	}

	/**
	 * サーバで例外が発生した際にクライアントでも例外を送出する場合はtrue、例外を送出しない場合はfalseが返る。
	 * <p>
	 * @return サーバで例外が発生した際にクライアントでも例外を送出する場合はtrue
	 */
	public boolean isThrowException() {
		return isThrowException;
	}

	/**
	 * サーバで例外が発生した際にクライアントでも例外を送出するかを設定する。
	 * <p>
	 * @param isThrowException サーバで例外が発生した際にクライアントでも例外を送出する場合はtrue
	 */
	public void setThrowException(boolean isThrowException) {
		this.isThrowException = isThrowException;
	}

	/**
	 * RCIサーバとのコネクションにより通信を行うInvokerクラスのインスタンスを取得する。
	 *<p>
	 * @return Invokerクラスのインスタンス
	 */
	public Invoker getInvoker() {
		return invoker;
	}

	/**
	 * サーバ側でクラス生成か参照の取得かstaticメソッドを遠隔実行する。
	 * <p>
	 * 第１パラメータが実行するクラスの種類を表しており、以下の３種類がある。<br>
	 * new: インスタンスクラスの生成<br>
	 * static: スタティッククラスのメソッドの実行<br>
	 * single: シングルトンクラスへの参照の取得<br>
	 * <br>
	 * 第２パラメータはクラスの完全修飾名を設定する。例)"java.util.HashMap"<br>
	 * <br>
	 * 第３パラメータ以降は第１パラメータにより以下になる。<br>
	 * new: コンストラクタのパラメータがあれば、それを与える。<br>
	 * static: メソッド名に続きパラメータがあれば、それを与える。<br>
	 * single: インスタンスを取得するメソッド名に続きパラメータがあれば、それを与える。<br>
	 * <br>
	 * 例)<br>
	 * new: Packet date = rci.invoke("new", "java.util.Date");<br>
	 * static: Packet result = rci.invoke("static", "java.lang.System", "getProperty","user.dir");<br>
	 * single: Packet singleton = rci.invoke("single", "example.Singleton", "getInstance");<br>
	 *
	 * @param objects 上記形式の可変パラメータ
	 * @return 実行結果のPacket
	 * @throws Exception 何らかの例外
	 */
	public Packet invoke(Object ... objects ) throws Exception {
		// サーバ側でクラス生成かstaticメソッドを実行する
		return invokeType(null, objects);
	}

	/**
	 * invokeのパラメータの前にパラメータの型を指定し、サーバ側でクラス生成か参照の取得かstaticメソッドを遠隔実行する。
	 * <p>
	 * パラメータの型を与えた方が適合するメソッドをより正確に見つけられる。
	 * @param paramTypes パラメータの型。例) intなら{int.class}のようにする。
	 * @param objects 可変パラメータ
	 * @return 実行結果のPacket
	 * @throws Exception 何らかの例外
	 */
	public Packet invokeType(Class<?>[] paramTypes, Object ... objects ) throws Exception {
		// サーバ側でクラス生成かstaticメソッドを実行する

		Packet req = new Packet();
		req.setParams(objects);
		if(paramTypes != null) {
			req.getOptionMap().put(Packet.OPTION_PARAMTYPES, paramTypes);
		}

		req.enSerializeParam();
		invoker.send(req);

		Packet res = invoker.recv();
		// スレッドリードならタイムアウトしたらnullが返る。タイムアウトしたらNullPointerExceptionになる。
		res.deSerializeParam();

		if(isThrowException) {
			if(res.getException() != null) {
				throw res.getException();
			}
		}
		for(int i = 0; i < objects.length; i++) {
			Packet.copyObject(res.getParams()[i], objects[i]);
		}
		res.setParams(objects);

		return res;
	}

	/**
	 * 鍵の交換を行いログインする。
	 * <p>
	 * @see com.graveng.rci.crypt.HashCrypt
	 * @param username ユーザ名
	 * @param password パスワード
	 * @param hashAlgorithm ハッシュアルゴリズム名(MD2, MD5, SHA-1, SHA-256, SHA-384, SHA-512の何れか)
	 * @param hashRepeat パラメータとして与えられたパスワードに対して更にハッシュ化を繰り返す数
	 * @throws Exception 何らかの例外
	 */
	public void authUser(String username, String password, String hashAlgorithm, int hashRepeat) throws Exception {
		invokeExchangeKey();
		invokeCertificateUser(username, password, hashAlgorithm, hashRepeat);
		return;
	}

	/**
	 * 公開鍵と秘密鍵に独自に生成したものを設定する。（このメソッドを呼び出さなければ自動生成したものが使われる。）
	 * <p>
	 * サーバ側で公開鍵が登録されていると、ここで与えた公開鍵と一致判定を行う。公開鍵と秘密鍵はRSAKeyPairCryptで生成する。
	 * @see com.graveng.rci.crypt.RSAKeyPairCrypt
	 * @param publicKey 公開鍵
	 * @param secretKey 秘密鍵
	 */
	public void setCryptKeyPair(String publicKey, String secretKey) {
		invoker.cryptFilter.pairCrypt.setPublicKey(publicKey);
		invoker.cryptFilter.pairCrypt.setSecretKey(secretKey);
	}

	/**
	 * 公開鍵をサーバに送信し、共通鍵を受け取る。
	 * <p>
	 * @return 実行結果のPacket
	 * @throws Exception 何らかの例外
	 */
	public Packet invokeExchangeKey() throws Exception {
		String publicKey = invoker.cryptFilter.getPublicKey();
		Packet res = invoke("ExchangeKey", publicKey);
		String[] keys = (String[])res.getResult();

		for(int i = 0; i < keys.length; i=i+2) {
			byte[] key = Base64.getDecoder().decode(keys[i].getBytes());
			byte[] iv = Base64.getDecoder().decode(keys[i+1].getBytes());
			AesCrypt ac = new AesCrypt();
			ac.setSecretKey(invoker.cryptFilter.pairCrypt.decryptBySecretKey(key));
			ac.setIvs(invoker.cryptFilter.pairCrypt.decryptBySecretKey(iv));
			invoker.cryptFilter.addAesCrypt(ac);
		}

		invoker.cryptFilter.setAesCryptMode(true);
		Logger.info("exchange key successfully");

		return res;
	}

	/**
	 * ログインする。
	 * <p>
	 * @param username ユーザ名
	 * @param password パスワード
	 * @param hashAlgorithm ハッシュアルゴリズム名
	 * @param hashRepeat パラメータとして与えられたパスワードに対して更にハッシュ化を繰り返す数
	 * @return 実行結果のPacket
	 * @throws Exception 何らかの例外
	 */
	public Packet invokeCertificateUser(String username, String password, String hashAlgorithm, int hashRepeat) throws Exception {
		String hashPassword;
		HashCrypt hc = new HashCrypt();
		hashPassword = hc.toEncryptedHashValue(hashAlgorithm, password, hashRepeat);

		Packet res = invoke("CertificateUser", username, hashPassword);
		Logger.info("successfully logged in user["+username+"]");
		return res;
	}

	/**
	 * パスワードを変更する。
	 * <p>
	 * @param hashAlgorithmName ハッシュアルゴリズム名
	 * @param loginPassword ログイン時のパスワード
	 * @param registPassword 登録するパスワード
	 * @param loginCount ログイン時のハッシュ繰り返し回数
	 * @param registCount 登録するパスワードのハッシュ繰り返し回数
	 * @param pubKey 公開鍵
	 * @return 実行結果のPacket
	 * @throws Exception 何らかの例外
	 */
	public Packet invokeUpdatePassword(String hashAlgorithmName, String loginPassword, String registPassword, int loginCount, int registCount, String pubKey) throws Exception {
		Packet res = invoke("UpdatePassword", hashAlgorithmName, loginPassword, registPassword, loginCount, registCount, pubKey );
		Logger.info("successfully update password");
		return res;
	}

	/**
	 * invokeで生成したインスタンス(かシングルトン)のメソッドを実行する。
	 * <p>
	 * 例)<br>
	 * new: date = rci.exec(date, "getMonth");<br>
	 * single: singleton = rci.exec(singleton, "setNameAndAge", "Yamada", 21);<br>
	 * <br>
	 * インスタンスの破棄はメソッド名以降なしで実行する。
	 * <p>
	 * 例)<br>
	 * rci.exec(date);<br>
	 * rci.exec(singleton);<br>
	 * <br>
	 * @param result invokeかinvokeTypeの返却値のPacket
	 * @param objects メソッド名、その後はパラメータ
	 * @return 実行結果Packet
	 * @throws Exception 何らかの例外
	 */
	public Packet exec(Packet result, Object ... objects ) throws Exception {
		return execType(result, null, objects);
	}

	/**
	 * パラメータの型を指定してinvokeで生成したインスタンス(かシングルトン)のメソッドを実行する。
	 * <p>
	 * @param result invokeかinvokeTypeの返却値のPacket
	 * @param paramTypes パラメータの型
	 * @param objects メソッド名、その後はパラメータ
	 * @return 実行結果Packet
	 * @throws Exception 何らかの例外
	 */
	public Packet execType(Packet result, Class<?>[] paramTypes, Object ... objects ) throws Exception {
		if(objects.length == 0) {
			// インスタンス破棄の場合
			Object lNumber = result.getOptionMap().get(Packet.OPTION_LISTENER_NUMBER);
			if(lNumber != null) {
				listenerMap.remove(lNumber);
			}
		}
		// サーバ側でメソッドを実行する

		Packet req = new Packet();
		req.setParams(objects);
		req.setInstanceNumber(result.getInstanceNumber());

		// リスナー情報は引き継ぐ
		req.getOptionMap().put(Packet.OPTION_LISTENER_NUMBER, result.getOptionMap().get(Packet.OPTION_LISTENER_NUMBER));
		req.getOptionMap().put(Packet.OPTION_LISTENER_HASHCODE, result.getOptionMap().get(Packet.OPTION_LISTENER_HASHCODE));

		if(paramTypes != null) {
			req.getOptionMap().put(Packet.OPTION_PARAMTYPES, paramTypes);
		}

		req.enSerializeParam();
		invoker.send(req);

		Packet res = invoker.recv();
		res.deSerializeParam();

		if(isThrowException) {
			if(res.getException() != null) {
				throw res.getException();
			}
		}
		for(int i = 0; i < objects.length; i++) {
			Packet.copyObject(res.getParams()[i], objects[i]);
		}
		res.setParams(objects);

		return res;
	}

	/**
	 * サーバ側でリスナー登録し、そのリスナーからの通知を受け取るリスナーを登録する。
	 * <p>
	 * @param result invokeかinvokeTypeの返却値のPacket
	 * @param clientRunListener クライアント側でコールバックされるリスナー
	 * @param objects メソッド名、その後はパラメータ
	 * @return 実行結果Packet
	 * @throws Exception 何らかの例外
	 */
	public Packet registListener(Packet result, EventListener clientRunListener, Object ... objects ) throws Exception {
		return registListener(result, null, clientRunListener, objects);
	}

	/**
	 * パラメータの型を指定し、サーバ側でリスナー登録し、そのリスナーからの通知を受け取るクラスを登録する。
	 * @param result invokeかinvokeTypeの返却値のPacket
	 * @param paramTypes パラメータの型
	 * @param clientRunListener クライアント側でコールバックされるリスナー
	 * @param objects メソッド名、その後はパラメータ
	 * @return 実行結果Packet
	 * @throws Exception 何らかの例外
	 */
	public Packet registListener(Packet result, Class<?>[] paramTypes, EventListener clientRunListener, Object ... objects ) throws Exception {
		// サーバ側でメソッドを実行する

		Packet req = new Packet();
		req.setParams(objects);
		req.setInstanceNumber(result.getInstanceNumber());

		if(paramTypes != null) {
			req.getOptionMap().put(Packet.OPTION_PARAMTYPES, paramTypes);
		}

		// listener処理
		boolean findListener = false;
		for(int i = 0; i < objects.length; i++) {
			if(objects[i] instanceof EventListener) {
				findListener = true;
			}
		}
		if(!findListener) {
			throw new IllegalArgumentException("Not found EventListener in objects");
		}
		invoker.setListener(new EventReceiver());


		int lNumber = ++listenerNumber;
		listenerMap.put(lNumber, clientRunListener);
		req.getOptionMap().put(Packet.OPTION_LISTENER_NUMBER, lNumber);

		req.enSerializeParam();
		invoker.send(req);

		Packet res = invoker.recv();
		res.deSerializeParam();

		if(isThrowException) {
			if(res.getException() != null) {
				throw res.getException();
			}
		}
		for(int i = 0; i < objects.length; i++) {
			Packet.copyObject(res.getParams()[i], objects[i]);
		}
		res.setParams(objects);

		return res;
	}

	/**
	 * ソケットを切断する。
	 * <p>
	 * @throws IOException 通信例外
	 */
	public void disconnect() throws IOException {
		if(socket != null) { socket.close(); }
	}

//	 * @throws InstantiationException
//	 * @throws IllegalAccessException
//	 * @throws ClassNotFoundException
//	 * @throws NoSuchMethodException
//	 * @throws SecurityException
//	 * @throws IllegalArgumentException
//	 * @throws InvocationTargetException
	/**
	 * サーバからの非同期通知に対応するリスナークラスをクライアントで実行する。
	 * <p>
	 * @param req サーバから受け取ったPacket
	 * @throws Exception 何らかの例外
	 */
	protected void invokeListener(Packet req) throws Exception {
		Object[] params;

		params = req.getParams();
		if(params.length == 0) {
			// インスタンスの破棄
			listenerMap.remove(req.getInstanceNumber());
			req.setInstanceNumber(0);
			return;
		}

		Object result;
		String methodName = params[0].toString();

		Object object = listenerMap.get(req.getInstanceNumber());
		if(object == null) {
			throw new ClassNotFoundException("not found instancee number:"+req.getInstanceNumber());
		}
		Class<? extends Object> c = object.getClass();

		int offset = 1;
		Object[] args = new Object[params.length -offset];
		Class<?>[] paramType = new Class[params.length -offset];
		for(int i = 0; i < paramType.length; i++) {
//			paramType[i] = params[i+offset].getClass();
			args[i] = params[i+offset];
			if(args[i] instanceof NullObject) {
				args[i] = null;
			}
//			paramType[i] = Packet.convertPrimitive(params[i+offset]);
			paramType[i] = Packet.convertPrimitive(args[i]);
		}

		Method method = null;
		try {
			// interfaceをインプリしていても適合だと判断されない。
			method = c.getDeclaredMethod(methodName, paramType);
		}
		catch(NoSuchMethodException e) {
			Logger.debug("NoSuchMethodException, so finding other way", e);
			Method[] methods = c.getDeclaredMethods();
			for(int i = 0; i < methods.length; i++) {
				if(methods[i].getName().equals(methodName) && methods[i].getParameterTypes().length == args.length ) {
					// メソッド名とパラメータ数が一致すれば適合と見なす
					method = methods[i];
					break;
				}
			}
		}

		result = method.invoke(object, args);
//		req.setInstance(object);
		req.setResult(result);
	}

	/**
	 * サーバからの非同期通知を受け取るクラス
	 */
	protected class EventReceiver implements DataReceivedListener {
		@Override
		public boolean received(Packet req) {
			// listenerの実行判定
			try {
				if(req.getOptionMap().get(Packet.OPTION_CALLBACK) != null) {
					req.deSerializeParam();
					invokeListener(req);
					return true;
				}
			} catch (Exception e) {
				Logger.warn("failed listener callback", e);
				return false;
			}
			return false;
		}
	}

	/**
	 * １つのコネクションの送受信を行う(Rci内部で使用するクラス)
	 */
	public class Invoker implements Runnable{
		protected Socket socket;
		protected DataInputStream dis = null;
		protected DataOutputStream dos = null;
		protected LinkedList<Packet> list = new LinkedList<Packet>();
		protected DataReceivedListener listener = null;
		protected boolean isThreadRead = false;

		/**
		 * ミリ秒の応答待ち１回のスリープ時間
		 */
		protected long recvSleepMs = 10;

		/**
		 * 応答待ち繰り返し最大数(この数を超えるとTimeoutExceptionを投げる)
		 */
		protected int recvSleepRetryCount = 10000;

		/**
		 * 待ち回数が増えるごとにスリープ時間を増やす(待ち回数をスリープ時間に掛ける)場合はtrue
		 * （早い応答には速やかに対応し、遅い応答にはCPUをあまり使わないようにする。）
		 */
		protected boolean isIncrementalSleep = true; //

		/**
		 * 暗号化するか平文にするかを判定するフィルター
		 */
		protected CryptFilter cryptFilter = new CryptFilter();

		public Invoker(Socket socket) {
			this.socket = socket;
			init();
		}

		/**
		 * ミリ秒の応答待ち１回のスリープ時間を取得する。
		 * @return ミリ秒の応答待ち１回のスリープ時間
		 */
		public long getRecvSleepMs() {
			return recvSleepMs;
		}

		/**
		 * ミリ秒の応答待ち１回のスリープ時間を設定する。
		 * @param recvSleepMs ミリ秒の応答待ち１回のスリープ時間
		 */
		public void setRecvSleepMs(long recvSleepMs) {
			this.recvSleepMs = recvSleepMs;
		}

		/**
		 * 応答待ち繰り返し最大数を取得する。
		 * @return 応答待ち繰り返し最大数
		 */
		public int getRecvSleepRetryCount() {
			return recvSleepRetryCount;
		}

		/**
		 * 応答待ち繰り返し最大数を設定する。
		 * @param recvSleepRetryCount 応答待ち繰り返し最大数
		 */
		public void setRecvSleepRetryCount(int recvSleepRetryCount) {
			this.recvSleepRetryCount = recvSleepRetryCount;
		}

		/**
		 *  待ち回数が増えるごとにスリープ時間を増やすかを取得する。
		 * @return 待ち回数が増えるごとにスリープ時間を増やすか
		 */
		public boolean isIncrementalSleep() {
			return isIncrementalSleep;
		}

		/**
		 * 待ち回数が増えるごとにスリープ時間を増やすかを設定する。
		 * @param isIncrementalSleep 待ち回数が増えるごとにスリープ時間を増やすか
		 */
		public void setIncrementalSleep(boolean isIncrementalSleep) {
			this.isIncrementalSleep = isIncrementalSleep;
		}

		/**
		 * 非同期通知に対応するためスレッドでソケットを読み込むかを取得する。
		 * @return スレッドでソケットを読み込む場合はtrue、送信と受信を交互に行う場合はfalse
		 */
		public boolean isThreadRead() {
			return isThreadRead;
		}

		public DataReceivedListener getListener() {
			return listener;
		}

		public void setListener(DataReceivedListener listener) {
			this.listener = listener;
		}

		public void close() {
			try {
	            if(dos != null) dos.close();
	            if(dis != null) dis.close();
	            if(socket != null) socket.close();
			}
			catch(Exception ie) {
				Logger.warn("close error", ie);
			}
		}

		protected void init() {
			try {
		        InputStream is = socket.getInputStream();
		        dis = new DataInputStream(is);

		        OutputStream os = socket.getOutputStream();
		        dos = new DataOutputStream(os);
			}
			catch(Exception e) {
				Logger.warn("initialize failure", e);
			}
		}

		public void run() {
			isThreadRead = true;
			try {
	    		while(true) {
		            // 入力処理
	    			synchronized(this) {
		    			Packet req = recvBlock();
			            recieved(req);
	    			}
	    		}
			}
			catch(Exception e) {
				Logger.info("communicate disconnect", e);
			}
			finally {
				close();
			}
		}

		// 要求1, 要求2に対し、応答2, 応答1と処理されるスレッドなら、キューも意味があるが、
		// 現状、要求1-応答1、要求2-応答2のようにスレッドではなく順番に処理するのでキューは意味がない。
		// 非同期通知を常に受けられるようにするために通常の要求応答と混在させるためキューを使っている。
		// 非同期の要求と応答を実現するには要求電文と応答電文を関連付ける情報(requestId)がいる。
		// ListではなくrequestIdをkeyにしたMapにすれば非同期処理も可能になるが、現状、対応していない。
		// (特に必要性が無く複雑なことをやっても有難迷惑かも知れないのでやっていない。)
		//（現状、RciServerも同じセッションの非同期の要求に対応していない。）
		//（httpdはリクエストを非同期で処理するが、１つのソケットのコネクション毎に非同期ということなので
		// 同じようにRciServerもコネクション毎では非同期に動いている。）
		protected Packet recvQueue() throws IOException {
			if(list.size() > 0) {
				Packet data = list.removeFirst();
				return data;
			}
			return null;
		}

		protected Packet recvQueue(long sleepTime, int count, boolean isIncrementalSleep) throws IOException {
			for(int i = 0; i < count; i++) {
				if(list.size() > 0) {
					Packet data = list.removeFirst();
					return data;
				}
				try {
					if(isIncrementalSleep) {
						Thread.sleep(sleepTime*i);
					}
					else {
						Thread.sleep(sleepTime);
					}
				} catch (InterruptedException e) {
//					e.printStackTrace();
					break;
				}
			}
			throw new TimeoutException("response timeout or interrupt");
//			return null;
		}

		protected void recieved(Packet res) throws IOException {
			if(listener != null) {
				if(listener.received(res)) {
					return;
				}
			}
			list.add(res);
		}

		protected Packet recvBlock() throws IOException {
			// 受信パケットを解析し、クラス生成やメソッド実行を行う。
			String text = dis.readUTF();
			text = cryptFilter.decrypt(text);
			Packet res = (Packet)Packet.getDecodeXML(text);
	        return  res;
		}

		protected Packet recv() throws IOException {
			if(!isThreadRead) {
				return recvBlock();
			}
			return recvQueue(recvSleepMs, recvSleepRetryCount, isIncrementalSleep);
		}

		protected void send(Packet req) throws IOException {
			// 応答を送信する。
			String sendData = Packet.getEncodeXML(req);
			sendData = cryptFilter.encrypt(sendData);
	        dos.writeUTF(sendData);
		}
	}

}

/**
 * 非同期イベント受信用のインタフェース
 *
 */
interface DataReceivedListener extends EventListener {
	boolean received(Packet text);
}