/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.crypt;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Base64;

/**
 * ハッシュ化の暗号化を行う。
 * <p>
 * ハッシュ化は片方向の暗号化の機能があり、パスワードの認証等に用いられている。
 * また、ファイルの改ざんや破壊の検知のためにも利用されている。
 * <br>
 * 最初の文字列のハッシュ化を特定回数繰り返すとそのハッシュ値はいつも同じ値になる。
 * この特徴を利用して、総当たり攻撃の対策をするために、ハッシュ化を繰り返せるようにしている。
 * <br>
 * ハッシュアルゴリズムは以下の何れかを利用できる。<br>
 * MD2, MD5, SHA-1, SHA-256, SHA-384, SHA-512<br>
 */
public class HashCrypt {
	/**
	 * ハッシュを繰り返すことで総当たり攻撃の対策をする。
	 * ただし、回数が多すぎると負荷要因にもなるので最大値を定義している。
	 */
	protected static final int MAX_HASH_REPEAT = 1000;

	/**
	 * ハッシュアルゴリズムのデフォルト値。以下のいずれかを利用できる。
	 * 	MD2, MD5, SHA-1, SHA-256, SHA-384, SHA-512
	 */
	protected String defaultAlgorithmName = "SHA-256";

	/**
	 * ハッシュ化を繰り返す回数のデフォルト値
	 */
	protected int repeatCount = 1;

	/**
	 * ２つのハッシュ化されたパスワードの照合(認証)を行う。
	 * <p>
	 * @param hashAlgorithmName ハッシュアルゴリズム名
	 * @param loginHashCount ログイン時のハッシュ化回数
	 * @param loginPassword ログイン時のパスワード
	 * @param registHashCount 登録されているパスワードのハッシュ化回数
	 * @param registPassword 登録されているパスワード
	 * @return trueの場合はログイン可能
	 */
	public boolean certificate(String hashAlgorithmName, int loginHashCount, String loginPassword, int registHashCount, String registPassword ) {
		int roundCount = loginHashCount - registHashCount;
    	if(roundCount >= 0) {
        	String password = toEncryptedHashValue(hashAlgorithmName, registPassword, roundCount);
        	if(!loginPassword.equals(password)) {
        		return false;
        	}
    	}
    	else {
    		roundCount = roundCount * (-1);
        	String password = toEncryptedHashValue(hashAlgorithmName, loginPassword, roundCount);
        	if(!registPassword.equals(password)) {
        		return false;
        	}
    	}
    	return true;
	}

	/**
	 * デフォルトの繰り返し回数とハッシュアルゴリズムでハッシュ化を行う。
	 * @param value ハッシュ化前の文字列
	 * @return ハッシュ化後の文字列
	 */
	public String toEncryptedHashValue(String value) {
		return toEncryptedHashValue(defaultAlgorithmName, value, repeatCount);
	}

	/**
	 * 文字列のハッシュ化を行う。
	 * @param algorithmName ハッシュアルゴリズム名
	 * @param value ハッシュ化前の文字列
	 * @param repeatCount ハッシュ化繰り返し回数
	 * @return ハッシュ化後の文字列
	 */
	public String toEncryptedHashValue(String algorithmName, String value, int repeatCount) {
		int repeat = repeatCount%MAX_HASH_REPEAT;
		String val = value;
		for(int i = 0; i < repeat; i++) {
			val = toEncryptedHashValue(algorithmName, val);
		}
		return val;
	}

	/**
	 * バイナリのハッシュ化を行う。
	 * @param algorithmName ハッシュアルゴリズム名
	 * @param value ハッシュ化前のバイナリ
	 * @param repeatCount ハッシュ化繰り返し回数
	 * @return ハッシュ化後のバイナリ
	 */
	public byte[] toEncryptedHashValue(String algorithmName, byte[] value, int repeatCount) {
		int repeat = repeatCount%MAX_HASH_REPEAT;
		byte[] val = value;
		for(int i = 0; i < repeat; i++) {
			val = toEncryptedHashValue(algorithmName, val);
		}
		return val;
	}

	/**
	 * 文字列のハッシュ化を行う。
	 * @param algorithmName ハッシュアルゴリズム名
	 * @param value ハッシュ化前の文字列
	 * @return ハッシュ化後の文字列
	 */
	public String toEncryptedHashValue(String algorithmName, String value) {
	    MessageDigest md = null;
	    StringBuilder sb = null;
	    try {
	        md = MessageDigest.getInstance(algorithmName);
	    } catch (NoSuchAlgorithmException e) {
	        throw new IllegalArgumentException(e);
	    }
	    md.update(value.getBytes());
	    sb = new StringBuilder();
	    for (byte b : md.digest()) {
	        String hex = String.format("%02x", b);
	        sb.append(hex);
	    }
	    return sb.toString();
	}

	/**
	 * バイナリのハッシュ化を行う。
	 * @param algorithmName ハッシュアルゴリズム名
	 * @param value ハッシュ化前のバイナリ
	 * @return ハッシュ化後のバイナリ
	 */
	public byte[] toEncryptedHashValue(String algorithmName, byte[] value) {
	    MessageDigest md = null;
	    try {
	        md = MessageDigest.getInstance(algorithmName);
	    } catch (NoSuchAlgorithmException e) {
	        throw new IllegalArgumentException(e);
	    }
	    md.update(value);

	    return md.digest();
	}

	/**
	 * ハッシュ化したものをBase64にして返す。
	 * @param algorithmName ハッシュアルゴリズム名
	 * @param text ハッシュ化前の文字列
	 * @return ハッシュ化後のBase64化した文字列
	 */
	public String toEncryptedHashBase64Value(String algorithmName, String text) {
		byte[] hashed = toEncryptedHashValue(algorithmName, text.getBytes());
		return toBase64String(hashed);
	}

	/**
	 * 16進数文字列形式に変換する。
	 * @param value バイナリ
	 * @return 16進数文字列形式
	 */
	public String toHexString(byte[] value) {
	    StringBuilder sb = new StringBuilder();
		for(int i = 0; i < value.length; i++) {
	        String hex = String.format("%02x", value[i]);
	        sb.append(hex);
	    }
	    return sb.toString();
	}

	/**
	 * Base64に変換する。
	 * @param value 変換前
	 * @return 変換後
	 */
	protected String toBase64String(byte[] value) {
    	return new String(Base64.getEncoder().encode(value));
	}

	/**
	 * デフォルトハッシュアルゴリズム名を取得する。
	 * @return デフォルトハッシュアルゴリズム名
	 */
	public String getDefaultAlgorithmName() {
		return defaultAlgorithmName;
	}

	/**
	 * デフォルトハッシュアルゴリズム名を設定する。
	 * @param defaultAlgorithmName デフォルトハッシュアルゴリズム名
	 */
	public void setDefaultAlgorithmName(String defaultAlgorithmName) {
		this.defaultAlgorithmName = defaultAlgorithmName;
	}

	/**
	 * デフォルトのハッシュ化繰り返し回数を取得する。
	 * @return デフォルトのハッシュ化繰り返し回数
	 */
	public int getRepeatCount() {
		return repeatCount;
	}

	/**
	 * デフォルトのハッシュ化繰り返し回数を設定する。
	 * @param repeatCount デフォルトのハッシュ化繰り返し回数
	 */
	public void setRepeatCount(int repeatCount) {
		this.repeatCount = repeatCount;
	}

	/**
	 * 文字列を指定回数分、文字列として繰り返したハッシュの文字列を出力する。
	 * （繰り返しにbyte[]を使うと文字列が16進数か判断する必要など複雑化するので単に文字列としてハッシュ化を繰り返す）
	 * （パスワードの初期のハッシュ暗号化時に利用する。）
	 *
	 * @param args
	 * args[0]:ハッシュ化する文字列（１６進数形式も単なる文字列としてハッシュ化する）
	 * args[1]:ハッシュ化の繰り返し回数
	 * args[2]:ハッシュアルゴリズム名（MD2, MD5, SHA-1, SHA-256, SHA-384, SHA-512のいずれか、省略時はSHA-256）
	 * args[3]:Base64化する場合はtrue, 16進数形式の場合はfalse、省略時はfalse
	 */
	public static void main(String[] args) {
		String usage = "usage:HashCrypt TEXT HASH_REPEAT_COUNT [ALGORITHM(MD2 | MD5 | SHA-1 | SHA-256 | SHA-384 | SHA-512)] [IS_BASE64(true|false)]";
		try {
			if(args.length < 2) {
				System.out.println(usage);
				System.exit(1);
			}
			boolean isBase64 = false;
			HashCrypt hc = new HashCrypt();
			String algo = "SHA-256";

			int max = Integer.parseInt(args[1]);
			String value = args[0];
			if(args.length >= 3) {
				algo = args[2];
			}
			if(args.length >= 4) {
				if("true".equals(args[3])) {
					isBase64 = true;
				}
			}
			for(int i = 0; i < max; i++) {
				if(isBase64) {
					value = hc.toEncryptedHashBase64Value(algo, value);
				}
				else {
					value = hc.toEncryptedHashValue(algo, value);
				}
			}
			System.out.println(value);
		}
		catch(Exception e) {
			System.out.println(usage);
			e.printStackTrace();
			System.exit(1);
		}
		System.exit(0);
	}
}
