/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.server;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;

import org.apache.commons.io.FileUtils;

import com.graveng.rci.server.config.Config;
import com.graveng.rci.server.config.GroupConfig;
import com.graveng.rci.server.config.IpAddressConfig;
import com.graveng.rci.server.config.ServerConfig;
import com.graveng.rci.server.config.UserConfig;
import com.graveng.rci.util.FileUtil;
import com.graveng.rci.util.Logger;
import com.graveng.rci.util.ObjectEx;
import com.graveng.rci.util.StringUtils;

/**
 * 設定ファイルからの情報へのアクセス口と設定ファイルの読み込みと保存を行う。(CmはConfigManagerのこと)
 * <p>
 * RciServerからアクセスされることを想定してパッケージプライベートのクラスとしている。<br>
 * com.graveng.rci.server.configの下のクラスはpublicであるが、単独で生成されても意味はなく、
 * このクラスで経由でアクセスする。
 *
 */
class Cm extends ObjectEx {
	/**
	 * 通常カレントディレクトリの下のdataに設定ファイルを置く。
	 */
	protected  String baseDirectory = System.getProperty("user.dir") + "/data";

	protected static final String ext = "xml";

	/**
	 * ディレクトリ形式ではなく、１つの設定ファイルに全部収録する場合の設定ファイル。(使われなくなった)
	 */
	protected static String fileName = "config.xml";

	/**
	 * ディレクトリで自分のクラス名の下に設定ファイルが置かれる
	 */
	protected String className = this.getClass().getSimpleName();

	/**
	 * シングルトンである
	 */
	protected static Cm own = new Cm();

	/**
	 * このConfigクラスの中に全設定情報が収録される。
	 */
	protected Config config = new Config(true);

	/**
	 * ディレクトリ形式の設定ファイルからロードする際、ユーザ名やグループ名はファイル名やディレクトリ名から取得し、
	 * HashMapのキーに設定するので、Config系クラスにロードする必要はないが、そこにもロードする場合はtrueにする。
	 */
	protected boolean isSetUser = true;

	/**
	 * 唯一のコンストラクタ
	 */
	protected Cm() {
		File file = new File(getFilePath());
		try {
			if(file.exists()) {
				// config.xmlからのロード（普通は使われない）
				load();
			}
			else {
				// ディレクトリに分散したxmlからロード
				loadDir();
			}
		} catch (Exception e) {
			String msg = "unable load config file";
			Logger.error(msg, e); // このログもおそらく出ない。

			// ここは処理の継続が不可能になるのでこんなところで終了せざるを得ない。
			System.out.println(msg);
			e.printStackTrace();
			System.exit(3);
		}
		initLog();
	}

	/**
	 * ログを初期化する。
	 */
	protected void initLog() {
		ServerConfig sc = config.getServerConfig();
		Logger.setLogFileName(sc.getLogFileName());
		Logger.setLogLevel(sc.getLogLevel());
	}

	/**
	 * インスタンスを取得する
	 * @return Cm
	 */
	public static Cm getInstance() {
		return own;
	}

	/**
	 * インスタンスを取得する
	 * @return Cm
	 */
	public static Cm gi() {
		return own;
	}

	/**
	 * 設定情報を取得する
	 * @return Config
	 */
	public Config getConfig() {
		return config;
	}

	/**
	 * config.xmlに一括して保存するもの
	 * @throws Exception 何らかの例外
	 */
	public void save() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();
		config.saveXML(getFilePath() );
	}

	/**
	 * ディレクトリ形式の個々の設定ファイルに保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveDir() throws Exception{
		saveServerDir();

		saveHostsAllowDir();
		saveHostsDenyDir();

		saveGroupDir();
		saveUserDir();
	}

	/**
	 * ServerConfig.xmlに保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveServerDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		ServerConfig sc = config.getServerConfig();
		sc.saveXML(getDir() + "/" + sc.getClass().getSimpleName() + "." + ext);
	}

	/**
	 * hosts/allowの下に保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveHostsAllowDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		HashMap<String, IpAddressConfig> map = config.getIpAddressAllowFilter();
		String[] hostName = map.keySet().toArray(new String[0]);
		for(int i = 0; i < hostName.length; i++) {
			String filename;
			filename = hostName[i] + "." + ext;
			IpAddressConfig uc = map.get(hostName[i]);

			uc.saveXML(getHostsAllowDir() + "/" + filename);
		}
	}

	/**
	 * hosts/denyの下に保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveHostsDenyDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		HashMap<String, IpAddressConfig> map = config.getIpAddressDenyFilter();
		String[] hostName = map.keySet().toArray(new String[0]);
		for(int i = 0; i < hostName.length; i++) {
			String filename;
			filename = hostName[i] + "." + ext;
			IpAddressConfig uc = map.get(hostName[i]);

			uc.saveXML(getHostsDenyDir() + "/" + filename);
		}
	}

	/**
	 * userの下にユーザ情報を保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveUserDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		HashMap<String, UserConfig> map = config.getUserNameFilter();
		String[] userName = map.keySet().toArray(new String[0]);
		for(int i = 0; i < userName.length; i++) {
			String filename;
			filename = userName[i] + "." + ext;
			UserConfig uc = map.get(userName[i]);

			File group = new File(getUserDir() + "/" + uc.getGroup());
			group.mkdirs();

			uc.saveXML(group.getAbsolutePath() + "/" + filename);
		}
	}

	/**
	 * userの下の特定のユーザのファイルに保存する。
	 * @throws IOException IO例外
	 */
	public void saveUser(String username) throws IOException {
		HashMap<String, UserConfig> map = config.getUserNameFilter();
		String filename = username + "." + ext;
		UserConfig uc = map.get(username);

		File group = new File(getUserDir() + "/" + uc.getGroup());
		group.mkdirs();

		uc.saveXML(group.getAbsolutePath() + "/" + filename);
	}

	/**
	 * userの下の特定のユーザのファイルのバックアップを生成する。
	 * @throws IOException IO例外
	 */
	public void backupUserFile(String username) throws IOException {
		HashMap<String, UserConfig> map = config.getUserNameFilter();
		String filename = username + "." + ext;
		UserConfig uc = map.get(username);

		File group = new File(getUserDir() + "/" + uc.getGroup());
		String srcFile = group.getAbsolutePath() + "/" + filename;
		String destFile = group.getAbsolutePath() + "/" + filename + ".bak";

		FileUtils.copyFile(new File(srcFile), new File(destFile));
	}

	/**
	 * userの下にグループ情報を保存する。
	 * @throws Exception 何らかの例外
	 */
	public void saveGroupDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		HashMap<String, GroupConfig> map = config.getGroupFilter();
		String[] userName = map.keySet().toArray(new String[0]);
		for(int i = 0; i < userName.length; i++) {
			GroupConfig uc = map.get(userName[i]);

			File group = new File(getUserDir() + "/" + userName[i]);
			group.mkdirs();

			uc.saveXML(group.getAbsolutePath() + "/" + "." + ext);
		}
	}

	/**
	 * config.xmlをConfigクラスにロードする。
	 * @throws Exception 何らかの例外
	 */
	protected void load() throws Exception{
		config =(Config) Config.loadXML(getFilePath());
	}

	/**
	 * ディレクトリ形式の設定ファイルからロードする。
	 * @throws Exception 何らかの例外
	 */
	public void loadDir() throws Exception{
		config = new Config();
		loadServerDir();

		loadHostsAllowDir();
		loadHostsDenyDir();

		loadUserDir();
	}

	/**
	 * ServerConfigにロードする。
	 * @throws Exception 何らかの例外
	 */
	public void loadServerDir() throws Exception{
		File dir = new File(getDir());
		dir.mkdirs();

		ServerConfig sc = new ServerConfig();
		String filepath = getDir() + "/" + sc.getClass().getSimpleName() + "." + ext;
		if(FileUtil.isFile(filepath)) {
//			sc = (ServerConfig)ServerConfig.loadXML(filepath);
			sc.loadXMLOwn(filepath);
			config.setServerConfig(sc);
		}
	}

	/**
	 * host/allowをロードする。
	 * @throws Exception 何らかの例外
	 */
	public void loadHostsAllowDir() throws Exception{
		File directory = new File(getHostsAllowDir());
		File[] files = FileUtil.listFiles(directory);
		for(int i = 0; i < files.length; i++) {
			if(files[i].isFile()) {
				String name = files[i].getName();
				String[] part = name.split("\\.");
				String nameOnly = name.replaceAll("." + ext, "");
				if(part.length >= 1 && part[part.length-1].equals("xml")) {
					IpAddressConfig ic = (IpAddressConfig) IpAddressConfig.loadXML(files[i].getAbsolutePath());
					config.getIpAddressAllowFilter().put(nameOnly, ic);
				}
			}
		}
	}

	/**
	 * host/denyをロードする。
	 * @throws Exception 何らかの例外
	 */
	public void loadHostsDenyDir() throws Exception{
		File directory = new File(getHostsDenyDir());
		File[] files = FileUtil.listFiles(directory);
		for(int i = 0; i < files.length; i++) {
			if(files[i].isFile()) {
				String name = files[i].getName();
				String[] part = name.split("\\.");
				String nameOnly = name.replaceAll("." + ext, "");
				if(part.length >= 1 && part[part.length-1].equals("xml")) {
					IpAddressConfig ic = (IpAddressConfig) IpAddressConfig.loadXML(files[i].getAbsolutePath());
					config.getIpAddressDenyFilter().put(nameOnly, ic);
				}
			}
		}
	}

	/**
	 * userをロードする。
	 * @throws Exception 何らかの例外
	 */
	public void loadUserDir() throws Exception{
		File directory = new File(getUserDir());
		File[] files = FileUtil.listFiles(directory);
		for(int i = 0; i < files.length; i++) {
			if(files[i].isFile()) {
				String name = files[i].getName();
				String[] part = name.split("\\.");
				String nameOnly = name.replaceAll("." + ext, "");
				String groupName = files[i].getParent();
				groupName = groupName.replaceAll("\\\\", "/");
				String userDir = getUserDir().replaceAll("\\\\", "/");
				groupName = groupName.replaceAll(userDir + "/",  "");

				if(name.equals(".xml")) {
					// グループ定義
					GroupConfig uc = (GroupConfig) GroupConfig.loadXML(files[i].getAbsolutePath());
					if(isSetUser) {
						uc.setName(groupName);
					}
//					uc.setGroup(getParentName(groupName));
					uc.setGroup(StringUtils.getSeparatorParent(groupName, "/"));
					config.getGroupFilter().put(groupName, uc);
				}
				else if(part.length >= 1 && part[part.length-1].equals("xml")) {
					// ユーザ定義
					UserConfig uc = (UserConfig) UserConfig.loadXML(files[i].getAbsolutePath());
					if(isSetUser) {
						uc.setName(nameOnly);
					}
					uc.setGroup(groupName);
					config.getUserNameFilter().put(nameOnly, uc);
				}
			}
		}
	}

	/**
	 * config.xmlのパスを取得する。
	 * @return パス
	 */
	protected String getFilePath() {
		String path = getDir() +"/"+fileName;
		return path;
	}

	/**
	 * 設定ファイルのディレクトリを取得する。
	 * @return パス
	 */
	protected String getDir() {
		String path = getBaseDirectory()+"/"+className;
		return path;
	}

	/**
	 * userディレクトリを取得する。
	 * @return パス
	 */
	protected String getUserDir() {
		String path = getBaseDirectory()+"/"+className +"/user";
		return path;
	}

	/**
	 * host/allowディレクトリを取得する。
	 * @return パス
	 */
	protected String getHostsAllowDir() {
		String path = getBaseDirectory()+"/"+className +"/hosts/allow";
		return path;
	}

	/**
	 * host/denyディレクトリを取得する。
	 * @return パス
	 */
	protected String getHostsDenyDir() {
		String path = getBaseDirectory()+"/"+className +"/hosts/deny";
		return path;
	}

	/**
	 * カレントディレクトリを取得する。
	 * @return baseDirectory
	 */
	public  String getBaseDirectory() {
		return baseDirectory;
	}

//	public static void main(String[] args) {
//		// 最初のXMLファイルの生成に使う
//		try {
//			Cm.gi().save();
//			String text = ClassField.toDump(Cm.gi().getConfig());
//			System.out.println(text);
//
//		} catch (Exception e) {
//			Logger.error("unable save config file", e);
//		}
//	}
}

